// RawH264WMF.cpp: fichier projet principal.

#include "stdafx.h"

#include <Codecapi.h>
#include <iCodecapi.h>
#include <mferror.h>
#include <mfapi.h>
#include <mfmediaengine.h>
#include <mfreadwrite.h>

#include <sqlite3.h>

#include <algorithm>
#include <iostream>
#include <vector>

//Utility function to display standard Windows error message
static void GetErrorMessage(HRESULT hr)
{
  if (hr)
  {
    LPTSTR errorText = 0;
    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_IGNORE_INSERTS,  
                  0, hr,  MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
                  (LPTSTR)&errorText, 0, 0);
    if (!errorText)
      {System::Console::WriteLine("errorMessage unknown for code = <{0:X}>", (unsigned)hr);}
    else//if (errorText)
    {
      System::String^ errorMessage =  gcnew System::String(errorText);
      System::Console::WriteLine("errorMessage for code = <{0:X}> = <{1}>", (unsigned)hr, errorMessage);
      delete errorMessage;
      LocalFree(errorText);
    }//end if (errorText)
  }//end if (hr)
}
//end GetErrorMessage()

//estimate relevant H264 bit rate
unsigned __int32 GetBestBitRate(unsigned int width, unsigned int height, double fps)
{
  unsigned __int32 result = 0;
  unsigned int motionFactor = 4;
  double bitRate = (1.*width)*(1.*height)*(fps)*(motionFactor)*(0.07);
  bitRate = std::min(bitRate, 1.*(AVENC_H264V_MAX_MBBITS)*(1024*1024));
  result = System::Convert::ToUInt32(System::Math::Round(bitRate));
  return result;
}
//end GetBestBitRate()

__int64 ConvertTimeToWMF(System::TimeSpan time)
{
  __int64 result = 0;
  static const __int64 NanoSecondsPerWMFTick = 100;
  System::Decimal value_d =
      System::Decimal(1e6)*System::Decimal(time.Ticks)/System::Decimal(System::TimeSpan::TicksPerMillisecond)/
                           System::Decimal(NanoSecondsPerWMFTick);
  result = System::Decimal::ToInt64(System::Decimal::Round(value_d));
  return result;
}
//end ConvertTimeToWMF(System::TimeSpan time)

//store a H264 sample with sub buffers pointing to SPS, PPS, plus some lags
class H264SampleComponents
{
  public:
    H264SampleComponents(void)
                       :begin(0),end(0),spsBegin(0),spsEnd(0),ppsBegin(0),ppsEnd(0),picBegin(0),picEnd(0),pictureFlagI(false),pictureFlagP(false),pictureFlagB(false) {}
    H264SampleComponents(const unsigned char* begin, const unsigned char* end)
                       :begin(begin),end(end),spsBegin(0),spsEnd(0),ppsBegin(0),ppsEnd(0),picBegin(0),picEnd(0),pictureFlagI(false),pictureFlagP(false),pictureFlagB(false) {}
    ~H264SampleComponents() {}
  public:
    const unsigned char* begin;
    const unsigned char* end;
    const unsigned char* spsBegin;
    const unsigned char* spsEnd;
    const unsigned char* ppsBegin;
    const unsigned char* ppsEnd;
    std::vector<unsigned char> seis;
    const unsigned char* picBegin;
    const unsigned char* picEnd;
    bool pictureFlagI;
    bool pictureFlagP;
    bool pictureFlagB;
};
//end class H264SampleComponents

//extract some H264 information from raw H264 sample
class H264Context
{
  public:
    H264Context(void) {}
  public:
    bool analyzeSample(const unsigned char* sampleBegin, const unsigned char* sampleEnd, H264SampleComponents* outSampleComponents)
    {
      bool result = false;
      bool resultError = false;

      H264SampleComponents sampleComponents(sampleBegin, sampleEnd);

      const unsigned char  naluSeparator[]     = {0x0, 0x0, 0x0, 0x1};
      const unsigned int   naluSeparatorLength = sizeof(naluSeparator);
      const unsigned char* naluSeparatorEnd    = naluSeparator+naluSeparatorLength;
      const unsigned char  naluSubSeparator[]     = {0x0, 0x0, 0x1};
      const unsigned int   naluSubSeparatorLength = sizeof(naluSubSeparator);
      const unsigned char* naluSubSeparatorEnd    = naluSubSeparator+naluSubSeparatorLength;
      const unsigned char* rawDataBegin = sampleBegin;
      const unsigned char* rawDataEnd   = sampleEnd;

      const unsigned char* reader = std::search(rawDataBegin, rawDataEnd, naluSeparator, naluSeparatorEnd);
      bool stop = (reader == rawDataEnd);
      if (stop){//a full nalu has not been found, we can consume almost all bytes
      }//end if (stop)
      else//if (!stop)
      {
        reader += naluSeparatorLength;
        while(!stop)
        {
          const unsigned char* nextNalu = std::search(reader, rawDataEnd, naluSubSeparator, naluSubSeparatorEnd);
          const unsigned char* nextNaluContent = (nextNalu == rawDataEnd) ? rawDataEnd : nextNalu+naluSubSeparatorLength;
          if ((nextNalu>reader+1) && !(*(nextNalu-1)))//maybe it was a nalu separator and not a subnaluseparator
            --nextNalu;
          unsigned char naluType = (*reader)&0x1F;//0b11111
          if (naluType == 7)//sps
          {
            if (nextNalu != rawDataEnd)
            {
              sampleComponents.spsBegin = reader;
              sampleComponents.spsEnd = nextNalu;
            }//end if (nextNalu != rawDataEnd)
          }//end if (naluType == 7)//sps
          else if (naluType == 8)//pps
          {
            if (nextNalu != rawDataEnd)
            {
              sampleComponents.ppsBegin = reader;
              sampleComponents.ppsEnd = nextNalu;
            }//end if (nextNalu != rawDataEnd)
          }//end if (naluType == 8)//sps
          else if (naluType == 9)//Access unit delimiter
          {
            if (reader+1 < rawDataEnd)
            {
              unsigned char primaryPicType = (*++reader)>>5;
              switch(primaryPicType)
              {
                case 0: sampleComponents.pictureFlagI = true;  sampleComponents.pictureFlagP = false; sampleComponents.pictureFlagB = false; break;
                case 1: sampleComponents.pictureFlagI = true;  sampleComponents.pictureFlagP = true;  sampleComponents.pictureFlagB = false; break;
                case 2: sampleComponents.pictureFlagI = true;  sampleComponents.pictureFlagP = true;  sampleComponents.pictureFlagB = true;  break;
                case 3: sampleComponents.pictureFlagI = false; sampleComponents.pictureFlagP = false; sampleComponents.pictureFlagB = false; break;
                case 4: sampleComponents.pictureFlagI = false; sampleComponents.pictureFlagP = false; sampleComponents.pictureFlagB = false; break;
                case 5: sampleComponents.pictureFlagI = true;  sampleComponents.pictureFlagP = false; sampleComponents.pictureFlagB = false; break;
                case 6: sampleComponents.pictureFlagI = true;  sampleComponents.pictureFlagP = true;  sampleComponents.pictureFlagB = false; break;
                case 7: sampleComponents.pictureFlagI = true;  sampleComponents.pictureFlagP = true;  sampleComponents.pictureFlagB = true;  break;
                default: break;
              }//end switch(primaryPicType)
            }//end if (reader+1 < rawDataEnd)
          }//end if (naluType == 9)//Access unit delimiter
          else if (naluType == 6)//sei
          {
            if (nextNalu != rawDataEnd)
            {
              const unsigned char* seiBegin = reader;
              const unsigned char* seiEnd = nextNalu;
              sampleComponents.seis.insert(sampleComponents.seis.end(), seiBegin, seiEnd);
            }//end if (nextNalu != rawDataEnd)
          }//end if (naluType == 6)//sei
          else if (naluType == 1)//P
          {
            sampleComponents.pictureFlagI = false;
            sampleComponents.pictureFlagP = true;
            sampleComponents.pictureFlagB = false;
            //if (nextNalu != rawDataEnd)
            {
              nextNalu = rawDataEnd;
              nextNaluContent = rawDataEnd;
              sampleComponents.picBegin = reader;
              sampleComponents.picEnd = nextNalu;
            }//end if (nextNalu != rawDataEnd)
          }//end if (naluType == 1)//P
          else if (naluType == 5)//I
          {
            sampleComponents.pictureFlagI = true;
            sampleComponents.pictureFlagP = false;
            sampleComponents.pictureFlagB = false;
            //if (nextNalu != rawDataEnd)
            {
              nextNalu = rawDataEnd;
              nextNaluContent = rawDataEnd;
              sampleComponents.picBegin = reader;
              sampleComponents.picEnd = nextNalu;
            }//end if (nextNalu != rawDataEnd)
          }//end if (naluType == 5)//I
          else
          {
            printf("unknown nalu %d\n", naluType);
          }
          reader = nextNaluContent;
          stop |= (reader == rawDataEnd);
        }//end while(!stop)

        bool hasSps = (sampleComponents.spsBegin != sampleComponents.spsEnd);
        bool hasPps = (sampleComponents.ppsBegin != sampleComponents.ppsEnd);
        bool hasSei = !sampleComponents.seis.empty();
        bool hasPic = (sampleComponents.picBegin != sampleComponents.picEnd);
        if (hasSps)
          this->lastSPS = std::vector<unsigned char>(sampleComponents.spsBegin, sampleComponents.spsEnd);
        if (hasPps)
          this->lastPPS = std::vector<unsigned char>(sampleComponents.ppsBegin, sampleComponents.ppsEnd);
        if (!hasSps)
        {
          sampleComponents.spsBegin = this->lastSPS.empty() ? 0 : &this->lastSPS[0];
          sampleComponents.spsEnd = sampleComponents.spsBegin+this->lastSPS.size();
          hasSps = (sampleComponents.spsBegin != sampleComponents.spsEnd);
        }//end if (!hasSps)
        if (!hasPps)
        {
          sampleComponents.ppsBegin = this->lastPPS.empty() ? 0 : &this->lastPPS[0];
          sampleComponents.ppsEnd = sampleComponents.ppsBegin+this->lastPPS.size();
          hasPps = (sampleComponents.ppsBegin != sampleComponents.ppsEnd);
        }//end if (!hasPps)
      }//end if (!stop)

      if (outSampleComponents)
        *outSampleComponents = sampleComponents;

      result = !resultError;
      return result;
    }//end analyzeH264Sample()
  public:
    std::vector<unsigned char> lastSPS;
    std::vector<unsigned char> lastPPS;
};
//end class analyzeSample

int main(array<System::String ^> ^args)
{
  HRESULT hr = S_OK;
  int error = 0;

  hr = CoInitializeEx(0, COINIT_MULTITHREADED);
  ::GetErrorMessage(hr);
  hr = MFStartup(MF_VERSION, 0);
  ::GetErrorMessage(hr);

  IMFAttributes* attributes = 0;
  hr = MFCreateAttributes(&attributes, 0);
  ::GetErrorMessage(hr);
  if (attributes)
  {
    //attributes->SetUINT32(MF_SINK_WRITER_DISABLE_THROTTLING, TRUE);
    attributes->SetGUID(MF_TRANSCODE_CONTAINERTYPE, MFTranscodeContainerType_MPEG4);
  }//end if (attributes)

  //First, we create the outputMediaType
  //The bytestream/mediasink/sinkwriter are delayed until we get the first sample with SPS and PPS
  unsigned int width = 640;
  unsigned int height = 480;
  double fps = 120;
  IMFMediaType* outputMediaType = 0;
  hr = MFCreateMediaType(&outputMediaType);
  ::GetErrorMessage(hr);
  if (outputMediaType)
  {
    hr = outputMediaType->SetGUID(MF_MT_MAJOR_TYPE, MFMediaType_Video);     
    ::GetErrorMessage(hr);
    hr = outputMediaType->SetGUID(MF_MT_SUBTYPE, MFVideoFormat_H264);   
    ::GetErrorMessage(hr);
    unsigned __int32 bitRate = ::GetBestBitRate(width, height, fps);
    hr = outputMediaType->SetUINT32(MF_MT_AVG_BITRATE, bitRate);
    ::GetErrorMessage(hr);
    hr = outputMediaType->SetUINT32(MF_MT_INTERLACE_MODE, MFVideoInterlace_Progressive);   
    ::GetErrorMessage(hr);
    hr = MFSetAttributeSize(outputMediaType, MF_MT_FRAME_SIZE, width, height);   
    ::GetErrorMessage(hr);
    std::pair<unsigned __int32, unsigned __int32> frameRate(120, 1);
    hr = MFSetAttributeRatio(outputMediaType, MF_MT_FRAME_RATE, frameRate.first, frameRate.second);
    ::GetErrorMessage(hr);
    hr = outputMediaType->SetUINT32(MF_MT_MPEG2_PROFILE, eAVEncH264VProfile_Main);
    ::GetErrorMessage(hr);
    //hr = outputMediaType->SetUINT32(MF_MT_MPEG2_LEVEL, eAVEncH264VLevel4_2);
    //::GetErrorMessage(hr);
    hr = MFSetAttributeRatio(outputMediaType, MF_MT_PIXEL_ASPECT_RATIO, 1, 1);   
    ::GetErrorMessage(hr);
    hr = outputMediaType->SetUINT32(MF_MT_MAX_KEYFRAME_SPACING, 16);
    ::GetErrorMessage(hr);
    hr = outputMediaType->SetUINT32(CODECAPI_AVEncCommonRateControlMode, eAVEncCommonRateControlMode_UnconstrainedVBR);//eAVEncCommonRateControlMode_Quality, eAVEncCommonRateControlMode_UnconstrainedCBR);
    ::GetErrorMessage(hr);
    hr = outputMediaType->SetUINT32(CODECAPI_AVEncCommonQuality, 100);
    ::GetErrorMessage(hr);
    hr = outputMediaType->SetUINT32(MF_MT_FIXED_SIZE_SAMPLES, FALSE);
    ::GetErrorMessage(hr);
    BOOL allSamplesIndependent = TRUE;
    hr = outputMediaType->SetUINT32(MF_MT_ALL_SAMPLES_INDEPENDENT, allSamplesIndependent);
    ::GetErrorMessage(hr);
    hr = outputMediaType->SetUINT32(MF_MT_COMPRESSED, TRUE);
    ::GetErrorMessage(hr);
  }//end if (outputMediaType)

  IMFByteStream* byteStream = 0;
  hr = MFCreateFile(MF_ACCESSMODE_READWRITE, MF_OPENMODE_DELETE_IF_EXIST, MF_FILEFLAGS_NONE, L"output.mp4", &byteStream);
  ::GetErrorMessage(hr);

  IMFMediaSink* mediaSink = 0;
  IMFMediaType* inputMediaType = 0;
  IMFSinkWriter* sinkWriter = 0;

  //we will also make raw output to process with ffmpeg
  //the command should be "ffmpeg.exe -f h264 -y -i output.h264 -vcodec copy -r 120/1 output-ffmpeg.mp4"
  FILE* rawH264OutputFile = fopen("output.h264", "wb");

  H264Context h264Context;

  //process samples from the SQLITE data base (ordered by time)
  error = 0;
  sqlite3* db = 0;
  error = sqlite3_open_v2("samples.db", &db, SQLITE_OPEN_READONLY, 0);
  static const char query[] = "SELECT time,duration,bytes from SAMPLES ORDER BY time";
  sqlite3_stmt* stmt = 0;
  error = sqlite3_prepare_v2(db, query, -1, &stmt, 0);
  unsigned int sampleCount = 0;
  bool stop = (error != SQLITE_OK);
  while(!stop)
  {
    System::Console::WriteLine("Process sample {0}", ++sampleCount);
    int status = sqlite3_step(stmt);
    stop |= (status != SQLITE_ROW);
    if (status == SQLITE_ROW)
    {
      int columnIndex = 0;
      __int64 timeTicks = sqlite3_column_int64(stmt, columnIndex++);//stored in System.TimeSpan ticks
      __int64 durationTicks = sqlite3_column_int64(stmt, columnIndex++);//stored in System.TimeSpan ticks
      int bytesLength = sqlite3_column_bytes(stmt, columnIndex);
      const void* bytes = sqlite3_column_blob(stmt, columnIndex);
      const unsigned char* sampleBegin = reinterpret_cast<const unsigned char*>(bytes);
      const unsigned char* sampleEnd = sampleBegin+bytesLength;

      //extract information like SPS, PPS and I flag
      H264SampleComponents sampleComponents(sampleBegin, sampleEnd);
      h264Context.analyzeSample(sampleBegin, sampleEnd, &sampleComponents);

      //the first sample should contain SPS and allow the initialization of MF_MT_MPEG_SEQUENCE_HEADER
      bool hasSPSPPS = ((sampleComponents.spsEnd-sampleComponents.spsBegin)>0) && ((sampleComponents.ppsEnd-sampleComponents.ppsBegin)>0);
      if (!mediaSink && outputMediaType && hasSPSPPS)
      {
        //the WMF docs tells how to build MF_MT_MPEG_SEQUENCE_HEADER by inserting the start code 0x0,0x0,0x1 between SPS and PPS
        static const unsigned char startCode[] = {0, 0, 1};
        unsigned int startCodeLength = sizeof(startCode);
        unsigned int spsLength = static_cast<unsigned int>(sampleComponents.spsEnd-sampleComponents.spsBegin);
        unsigned int ppsLength = static_cast<unsigned int>(sampleComponents.ppsEnd-sampleComponents.ppsBegin);
        unsigned int bufferLength = startCodeLength+spsLength+startCodeLength+ppsLength;
        unsigned char* buffer = (unsigned char*)calloc(bufferLength, sizeof(unsigned char));
        if (buffer)
        {
          unsigned int offset = 0;
          memcpy(buffer+offset, startCode, startCodeLength);
          offset += startCodeLength;
          memcpy(buffer+offset, sampleComponents.spsBegin, spsLength);
          offset += spsLength;
          memcpy(buffer+offset, startCode, startCodeLength);
          offset += startCodeLength;
          memcpy(buffer+offset, sampleComponents.ppsBegin, ppsLength);
          offset += ppsLength;
          HRESULT hr = S_OK;

          //now create the inputMediaType as a clone of the outputMediaType, since H264 is already encoded
          //we will just adjust MF_MT_MPEG_SEQUENCE_HEADER
          hr = MFCreateMediaType(&inputMediaType);
          ::GetErrorMessage(hr);
          if (inputMediaType)
          {
            hr = outputMediaType->CopyAllItems(inputMediaType);
            ::GetErrorMessage(hr);
            hr = inputMediaType->SetBlob(MF_MT_MPEG_SEQUENCE_HEADER, buffer, bufferLength);
            ::GetErrorMessage(hr);

            hr = MFCreateMPEG4MediaSink(byteStream, inputMediaType, 0, &mediaSink);
            ::GetErrorMessage(hr);
            if (mediaSink)
            {
              hr = MFCreateSinkWriterFromMediaSink(mediaSink, 0, &sinkWriter);
              ::GetErrorMessage(hr);
              if (sinkWriter)
              {
                //inputMediaType returns an error when set to MFVideoFormat_H264
                //I assume this is an dummy error and that it must be ignored
                hr = sinkWriter->SetInputMediaType(0, inputMediaType, 0);
                ::GetErrorMessage(hr);
                hr = sinkWriter->BeginWriting();
                ::GetErrorMessage(hr);
              }//end if (this->sinkWriter)
            }//end if (this->mediaSink)
          }//end if (inputMediaType)
          free(buffer);
        }//end if (buffer)
      }//end if (!mediaSink && outputMediaType && hasSPSPPS)

      //for all samples, we will just send them to the sinkwriter which has been properly initialized
      if (sinkWriter)
      {
        HRESULT hr = 0;
        IMFSample* sample = 0;
        IMFMediaBuffer* mediaBuffer = 0;

        unsigned int sampleSize = sampleEnd-sampleBegin;
        BYTE* pData = 0;
        hr = MFCreateAlignedMemoryBuffer(sampleSize, MF_16_BYTE_ALIGNMENT, &mediaBuffer);
        ::GetErrorMessage(hr);
        if (mediaBuffer)
        {
          hr = mediaBuffer->Lock(&pData, 0, 0);
          ::GetErrorMessage(hr);
          if (SUCCEEDED(hr))
          {
            memcpy(pData, sampleBegin, sampleSize);
            hr = mediaBuffer->Unlock();
            ::GetErrorMessage(hr);
          }//end if (SUCCEEDED(hr))
          hr = mediaBuffer->SetCurrentLength(sampleSize);
          ::GetErrorMessage(hr);
          hr = MFCreateSample(&sample);
          ::GetErrorMessage(hr);
          if (sample)
          {
            hr = sample->AddBuffer(mediaBuffer);
            ::GetErrorMessage(hr);
            __int64 mediaTimeStamp = ::ConvertTimeToWMF(System::TimeSpan(timeTicks));
            __int64 mediaDuration = ::ConvertTimeToWMF(System::TimeSpan(durationTicks));
            hr = sample->SetSampleTime(mediaTimeStamp);
            ::GetErrorMessage(hr);
            hr = sample->SetSampleDuration(mediaDuration);
            ::GetErrorMessage(hr);
            if (sampleComponents.pictureFlagI)
            {
              hr = sample->SetUINT32(MFSampleExtension_CleanPoint, TRUE);
              ::GetErrorMessage(hr);
            }//end if (sampleComponents.pictureFlagI)
            /*hr = MFGetAttributeUINT32(sample, MFSampleExtension_VideoEncodePictureType,
              sampleComponents.pictureFlagI ? eAVEncH264PictureType_IDR :
              sampleComponents.pictureFlagP ? eAVEncH264PictureType_P :
              sampleComponents.pictureFlagB ? eAVEncH264PictureType_B :
              eAVEncH264PictureType_IDR);
            ::GetErrorMessage(hr);*/
            hr = sinkWriter->WriteSample(0, sample);
            if (rawH264OutputFile)
              fwrite(sampleBegin, sizeof(unsigned char), sampleSize, rawH264OutputFile);
            ::GetErrorMessage(hr);
          }//end if (sample)
        }//end if (mediaBuffer)
        if (sample)
          sample->Release();
        if (mediaBuffer)
          mediaBuffer->Release();
      }//end if (sinkWriter)
    }//end if (status == SQLITE_ROW)
  }//end while(!stop)
  error = sqlite3_finalize(stmt);

  if (sinkWriter)
  {
    /*hr = sinkWriter->NotifyEndOfSegment(MF_SINK_WRITER_ALL_STREAMS);
    ::GetErrorMessage(hr);
    hr = sinkWriter->Flush(MF_SINK_WRITER_ALL_STREAMS);
    ::GetErrorMessage(hr);*/
    hr = sinkWriter->Finalize();
    ::GetErrorMessage(hr);
  }//end if (sinkWriter)

  if (inputMediaType)
    inputMediaType->Release();
  if (sinkWriter)
    sinkWriter->Release();
  if (mediaSink)
    mediaSink->Release();
  if (byteStream)
    byteStream->Release();

  if (rawH264OutputFile)
    fclose(rawH264OutputFile);

  MFShutdown();
  CoUninitialize();

  System::Console::WriteLine("Done");
  System::Console::ReadKey();

  return 0;
}
